/*
 * PropertiesListListener.java Copyright (C) 2020. Daniel H. Huson
 *
 * (Some code written by other authors, as named in code.)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package jloda.util;

import java.util.List;

/**
 * listen for changes to list of values of a property
 *
 * @author huson
 * Date: 11-Dec-2004
 */
public interface PropertiesListListener {
    /**
     * gets the new list of values after the change
     *
     * @param values
     */
    void hasChanged(List<String> values);

    /**
     * is this listener interesed in the named property?
     *
     * @param name
     * @return is interested
     */
    boolean isInterested(String name);
}
